#ifndef TOML11_DATETIME_HPP
#define TOML11_DATETIME_HPP

#include "fwd/datetime_fwd.hpp" // IWYU pragma: export

#if ! defined(TOML11_COMPILE_SOURCES)
#include "impl/datetime_impl.hpp" // IWYU pragma: export
#endif

#endif // TOML11_DATETIME_HPP
